#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright 2019 Daniel Estevez <daniel@destevez.net>
#
# This file is part of gr-satellites
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

from gnuradio import gr, analog, blocks, digital, filter
from gnuradio.filter import firdes

from math import ceil, pi
import pathlib
import sys

from ...hier.rms_agc import rms_agc
from ...utils.options_block import options_block
from ... import manchester_sync

class bpsk_demodulator(gr.hier_block2, options_block):
    """
    Hierarchical block to demodulate BPSK.

    The input can be either IQ or real.

    Args:
        baudrate: Baudrate in symbols per second (float)
        sample_rate: Sample rate in samples per second (float)
        iq: Whether the input is IQ or real (bool)
        f_offset: Frequency offset in Hz (float)
        differential: Perform non-coherent DBPSK decoding (bool)
        manchester: Use Manchester coding (bool)
        dump_path: Path to dump internal signals to files (str)
        options: Options from argparse
    """
    def __init__(self, baudrate, samp_rate, iq, f_offset = None, differential = False, manchester = False, dump_path = None, options = None):
        gr.hier_block2.__init__(self, "bpsk_demodulator",
            gr.io_signature(1, 1, gr.sizeof_gr_complex if iq else gr.sizeof_float),
            gr.io_signature(1, 1, gr.sizeof_float))
        options_block.__init__(self, options)

        if dump_path is not None:
            dump_path = pathlib.Path(dump_path)

        if manchester:
            baudrate *= 2

        # prevent problems due to baudrate too high
        if baudrate >= samp_rate / 4:
            print(f'Sample rate {samp_rate} sps insufficient for {baudrate} baud BPSK demodulation. Demodulator will not work.',
                      file = sys.stderr)
            baudrate = samp_rate / 4
            
        sps = samp_rate / baudrate
        max_sps = 10
        if sps > max_sps:
            decimation = ceil(sps / max_sps)
        else:
            decimation = 1
        sps /= decimation

        filter_cutoff = baudrate * 2.0
        filter_transition = baudrate * 0.2

        if f_offset is None:
            f_offset = self.options.f_offset
        if f_offset is None:
            if iq:
                f_offset = 0
            elif baudrate <= 2400:
                f_offset = 1500
            else:
                f_offset = 12000

        taps = firdes.low_pass(1, samp_rate, filter_cutoff, filter_transition)
        f = filter.freq_xlating_fir_filter_ccf if iq else filter.freq_xlating_fir_filter_fcf
        self.xlating = f(decimation, taps, f_offset, samp_rate)

        agc_constant = 2e-2 / sps # This gives a time constant of 50 symbols
        self.agc = rms_agc(agc_constant, 1)

        if dump_path is not None:
            self.agc_in = blocks.file_sink(gr.sizeof_gr_complex, str(dump_path / 'agc_in.c64'), False)
            self.agc_out = blocks.file_sink(gr.sizeof_gr_complex, str(dump_path / 'agc_out.c64'), False)
            self.connect(self.xlating, self.agc_in)
            self.connect(self.agc, self.agc_out)

        if not self.options.disable_fll:
            fll_bw = 2*pi*decimation/samp_rate*self.options.fll_bw
            self.fll = digital.fll_band_edge_cc(sps, self.options.rrc_alpha, 100, fll_bw)

            if dump_path is not None:
                self.fll_freq = blocks.file_sink(gr.sizeof_float, str(dump_path / 'fll_freq.f32'), False)
                self.fll_phase = blocks.file_sink(gr.sizeof_float, str(dump_path / 'fll_phase.f32'), False)
                self.fll_error = blocks.file_sink(gr.sizeof_float, str(dump_path / 'fll_error.f32'), False)
                self.connect((self.fll, 1), self.fll_freq)
                self.connect((self.fll, 2), self.fll_phase)
                self.connect((self.fll, 3), self.fll_error)            

        nfilts = 16
        rrc_taps = firdes.root_raised_cosine(nfilts, nfilts, 1.0/float(sps), self.options.rrc_alpha, int(ceil(11*sps*nfilts)))
        ted_gain = 0.5 # "empiric" formula for TED gain of a PFB MF TED for complex BPSK 0.5 sample^{-1}
        damping = 1.0
        self.clock_recovery = digital.symbol_sync_cc(digital.TED_SIGNAL_TIMES_SLOPE_ML,
                                                     sps,
                                                     self.options.clk_bw,
                                                     damping,
                                                     ted_gain,
                                                     self.options.clk_limit * sps,
                                                     1,
                                                     digital.constellation_bpsk().base(),
                                                     digital.IR_PFB_MF,
                                                     nfilts,
                                                     rrc_taps)
        
        if dump_path is not None:
            self.clock_recovery_out = blocks.file_sink(gr.sizeof_gr_complex, str(dump_path / 'clock_recovery_out.c64'), False)
            self.clock_recovery_err = blocks.file_sink(gr.sizeof_float, str(dump_path / 'clock_recovery_err.f32'), False)
            self.clock_recovery_T_inst = blocks.file_sink(gr.sizeof_float, str(dump_path / 'clock_recovery_T_inst.f32'), False)
            self.clock_recovery_T_avg = blocks.file_sink(gr.sizeof_float, str(dump_path / 'clock_recovery_T_avg.f32'), False)
            self.connect(self.clock_recovery, self.clock_recovery_out)
            self.connect((self.clock_recovery, 1), self.clock_recovery_err)
            self.connect((self.clock_recovery, 2), self.clock_recovery_T_inst)
            self.connect((self.clock_recovery, 3), self.clock_recovery_T_avg)

        self.connect(self, self.xlating, self.agc)
        if self.options.disable_fll:
            self.connect(self.agc, self.clock_recovery)
        else:
            self.connect(self.agc, self.fll, self.clock_recovery)

        self.complex_to_real = blocks.complex_to_real(1)

        if manchester:
            self.manchester = manchester_sync(self.options.manchester_history)
            self.connect(self.clock_recovery, self.manchester)
        else:
            self.manchester = self.clock_recovery
                
        if differential:
            self.delay = blocks.delay(gr.sizeof_gr_complex, 1)
            self.multiply_conj = blocks.multiply_conjugate_cc(1)
            sign = -1 if manchester else 1
            self.multiply_const = blocks.multiply_const_ff(sign, 1) # take care about inverion in Manchester
            self.connect(self.manchester, (self.multiply_conj, 0))
            self.connect(self.manchester, self.delay, (self.multiply_conj, 1))
            self.connect(self.multiply_conj, self.complex_to_real, self.multiply_const, self)
        else:
            costas_bw = 2*pi/baudrate*self.options.costas_bw
            self.costas = digital.costas_loop_cc(costas_bw, 2, False)

            if dump_path is not None:
                self.costas_out = blocks.file_sink(gr.sizeof_gr_complex, str(dump_path / 'costas_out.c64'), False)
                self.costas_frequency = blocks.file_sink(gr.sizeof_float, str(dump_path / 'costas_frequency.f32'), False)
                self.costas_phase = blocks.file_sink(gr.sizeof_float, str(dump_path / 'costas_phase.f32'), False)
                self.costas_error = blocks.file_sink(gr.sizeof_float, str(dump_path / 'costas_error.f32'), False)
                self.connect(self.costas, self.costas_out)
                self.connect((self.costas, 1), self.costas_frequency)
                self.connect((self.costas, 2), self.costas_phase)
                self.connect((self.costas, 3), self.costas_error)
            
            self.connect(self.manchester, self.costas, self.complex_to_real, self)

    _default_rrc_alpha = 0.35
    _default_fll_bw = 25
    _default_clk_rel_bw = 0.06
    _default_clk_limit = 0.004
    _default_costas_bw = 50
    _default_manchester_history = 32
    
    @classmethod
    def add_options(cls, parser):
        """
        Adds BPSK demodulator specific options to the argparse parser
        """
        parser.add_argument('--f_offset', type = float, default = None, help = 'Frequency offset (Hz) [default=1500 or 12000]')
        parser.add_argument('--rrc_alpha', type = float, default = cls._default_rrc_alpha, help = 'RRC roll-off (Hz) [default=%(default)r]')
        parser.add_argument('--disable_fll', action = 'store_true', help = 'Disable FLL')
        parser.add_argument('--fll_bw', type = float, default = cls._default_fll_bw, help = 'FLL bandwidth (Hz) [default=%(default)r]')
        parser.add_argument('--clk_bw', type = float, default = cls._default_clk_rel_bw, help = 'Clock recovery bandwidth (relative to baudrate) [default=%(default)r]')
        parser.add_argument('--clk_limit', type = float, default = cls._default_clk_limit, help = 'Clock recovery limit (relative to baudrate) [default=%(default)r]')
        parser.add_argument('--costas_bw', type = float, default = cls._default_costas_bw, help = 'Costas loop bandwidth (Hz) [default=%(default)r]')
        parser.add_argument('--manchester_history', type = int, default = cls._default_manchester_history, help = 'Manchester recovery history (symbols) [default=%(default)r]')



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright 2019 Daniel Estevez <daniel@destevez.net>
#
# This file is part of gr-satellites
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

import numpy
from gnuradio import gr
import pmt
from .check_eseo_crc import crc16_ccitt_zero as crc16_ccitt_zero

class check_swiatowid_crc(gr.basic_block):
    """
    docstring for block check_swiatowid_crc
    """
    def __init__(self, verbose):
        gr.basic_block.__init__(self,
            name="check_swiatowid_crc",
            in_sig=[],
            out_sig=[])

        self.verbose = verbose
        
        self.message_port_register_in(pmt.intern('in'))
        self.set_msg_handler(pmt.intern('in'), self.handle_msg)
        self.message_port_register_out(pmt.intern('ok'))
        self.message_port_register_out(pmt.intern('fail'))

    def handle_msg(self, msg_pmt):
        msg = pmt.cdr(msg_pmt)
        if not pmt.is_u8vector(msg):
            print("[ERROR] Received invalid message type. Expected u8vector")
            return
        packet = pmt.u8vector_elements(msg)

        if len(packet) < 3:
            return
        
        packet_out = packet[:-2]
        msg_out = pmt.cons(pmt.car(msg_pmt), pmt.init_u8vector(len(packet_out), packet_out))
        crc = crc16_ccitt_zero(packet_out)
        if packet[-2] == crc & 0xff and packet[-1] == crc >> 8:
            if self.verbose:
                print("CRC OK")
            self.message_port_pub(pmt.intern('ok'), msg_out)
        else:
            if self.verbose:
                print("CRC failed")
            self.message_port_pub(pmt.intern('fail'), msg_out)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright 2019 Daniel Estevez <daniel@destevez.net>
#
# This file is part of gr-satellites
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

from gnuradio import gr, filter, analog
from gnuradio.filter import firdes
from .fsk_demodulator import fsk_demodulator
from ...utils.options_block import options_block

class afsk_demodulator(gr.hier_block2, options_block):
    """
    Hierarchical block to demodulate AFSK.

    The input can be either IQ or real. For IQ input, it is assumed
    that the data is FM modulated, so FM demodulation is performed.
    For real input, it is assumed that the data is already FM
    demodulated.

    Args:
        baudrate: Baudrate in symbols per second (float)
        sample_rate: Sample rate in samples per second (float)
        iq: Whether the input is IQ or real (bool)
        af_carrier: Audio frequency carrier in Hz (float)
        deviation: Deviation in Hz, negative inverts the sidebands (float)
        dump_path: Path to dump internal signals to files (str)
        options: Options from argparse
    """
    def __init__(self, baudrate, samp_rate, iq, af_carrier, deviation, dump_path = None, options = None):
        gr.hier_block2.__init__(self, "afsk_demodulator",
            gr.io_signature(1, 1, gr.sizeof_gr_complex if iq else gr.sizeof_float),
            gr.io_signature(1, 1, gr.sizeof_float))
        options_block.__init__(self, options)
        
        if iq:
            self.demod = analog.quadrature_demod_cf(1)
            self.connect(self, self.demod)
        else:
            self.demod = self

        filter_cutoff = 2 * abs(deviation)
        filter_transition = 0.1 * abs(deviation)
        taps = firdes.low_pass(1, samp_rate, filter_cutoff, filter_transition)
        self.xlating = filter.freq_xlating_fir_filter_fcf(1, taps, af_carrier, samp_rate)

        self.fsk = fsk_demodulator(baudrate, samp_rate, deviation = deviation, iq = True,
                                    dc_block = False, dump_path = dump_path, options = options)

        self.connect(self.demod, self.xlating, self.fsk, self)

    @classmethod
    def add_options(cls, parser):
        """
        Adds CCSDS concatenated deframer specific options to the argparse parser
        """
        fsk_demodulator.add_options(parser)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright 2017,2020 Daniel Estevez <daniel@destevez.net>
#
# This file is part of gr-satellites
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

import numpy
from gnuradio import gr
import pmt

class sat_3cat_2:
    def parse(self, packet):
        data = packet[17:].split()
        
        status = {b'1' : 'Survival', b'2' : 'Sun-safe', b'3' : 'Nominal',\
                  b'4' : 'TX', b'5' : 'RX', b'6' : 'Payload', b'7' : 'Payload' }
        adcs = {b'0' : 'auto', b'1' : 'manual'}
        if data[5] == b'0':
            detumbling = 'Detumbling  ({},{},{})nT'.format(float(data[7]), float(data[8]), float(data[9]))
        else:
            detumbling = 'SS-nominal  Sun: ({:.2f},{:.2f},{:.2f})'.format(float(data[7]), float(data[8]), float(data[9]))

        string = status[data[0]] + '  {:.2f}V  {}mA'.format(int(data[1])/1000.0, int(data[2])) + \
        '  EPS: {}ºC   Ant: {}ºC'.format(data[3], data[4]) + \
        '  ADCS ' + adcs[data[6]] + '  ' + detumbling + \
        '  Control: ({:.1e},{:.1e},{:.1e})V'.format(float(data[10]), float(data[11]), float(data[12]))

        return string



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright 2021 Daniel Estevez <daniel@destevez.net>
#
# This file is part of gr-satellites
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

import numpy
from gnuradio import gr
import pmt

import struct

crc_table = [\
    0x0000, 0x1021, 0x2042, 0x3063, 0x4084, 0x50a5, 0x60c6, 0x70e7,\
    0x8108, 0x9129, 0xa14a, 0xb16b, 0xc18c, 0xd1ad, 0xe1ce, 0xf1ef,\
    0x1231, 0x0210, 0x3273, 0x2252, 0x52b5, 0x4294, 0x72f7, 0x62d6,\
    0x9339, 0x8318, 0xb37b, 0xa35a, 0xd3bd, 0xc39c, 0xf3ff, 0xe3de,\
    0x2462, 0x3443, 0x0420, 0x1401, 0x64e6, 0x74c7, 0x44a4, 0x5485,\
    0xa56a, 0xb54b, 0x8528, 0x9509, 0xe5ee, 0xf5cf, 0xc5ac, 0xd58d,\
    0x3653, 0x2672, 0x1611, 0x0630, 0x76d7, 0x66f6, 0x5695, 0x46b4,\
    0xb75b, 0xa77a, 0x9719, 0x8738, 0xf7df, 0xe7fe, 0xd79d, 0xc7bc,\
    0x48c4, 0x58e5, 0x6886, 0x78a7, 0x0840, 0x1861, 0x2802, 0x3823,\
    0xc9cc, 0xd9ed, 0xe98e, 0xf9af, 0x8948, 0x9969, 0xa90a, 0xb92b,\
    0x5af5, 0x4ad4, 0x7ab7, 0x6a96, 0x1a71, 0x0a50, 0x3a33, 0x2a12,\
    0xdbfd, 0xcbdc, 0xfbbf, 0xeb9e, 0x9b79, 0x8b58, 0xbb3b, 0xab1a,\
    0x6ca6, 0x7c87, 0x4ce4, 0x5cc5, 0x2c22, 0x3c03, 0x0c60, 0x1c41,\
    0xedae, 0xfd8f, 0xcdec, 0xddcd, 0xad2a, 0xbd0b, 0x8d68, 0x9d49,\
    0x7e97, 0x6eb6, 0x5ed5, 0x4ef4, 0x3e13, 0x2e32, 0x1e51, 0x0e70,\
    0xff9f, 0xefbe, 0xdfdd, 0xcffc, 0xbf1b, 0xaf3a, 0x9f59, 0x8f78,\
    0x9188, 0x81a9, 0xb1ca, 0xa1eb, 0xd10c, 0xc12d, 0xf14e, 0xe16f,\
    0x1080, 0x00a1, 0x30c2, 0x20e3, 0x5004, 0x4025, 0x7046, 0x6067,\
    0x83b9, 0x9398, 0xa3fb, 0xb3da, 0xc33d, 0xd31c, 0xe37f, 0xf35e,\
    0x02b1, 0x1290, 0x22f3, 0x32d2, 0x4235, 0x5214, 0x6277, 0x7256,\
    0xb5ea, 0xa5cb, 0x95a8, 0x8589, 0xf56e, 0xe54f, 0xd52c, 0xc50d,\
    0x34e2, 0x24c3, 0x14a0, 0x0481, 0x7466, 0x6447, 0x5424, 0x4405,\
    0xa7db, 0xb7fa, 0x8799, 0x97b8, 0xe75f, 0xf77e, 0xc71d, 0xd73c,\
    0x26d3, 0x36f2, 0x0691, 0x16b0, 0x6657, 0x7676, 0x4615, 0x5634,\
    0xd94c, 0xc96d, 0xf90e, 0xe92f, 0x99c8, 0x89e9, 0xb98a, 0xa9ab,\
    0x5844, 0x4865, 0x7806, 0x6827, 0x18c0, 0x08e1, 0x3882, 0x28a3,\
    0xcb7d, 0xdb5c, 0xeb3f, 0xfb1e, 0x8bf9, 0x9bd8, 0xabbb, 0xbb9a,\
    0x4a75, 0x5a54, 0x6a37, 0x7a16, 0x0af1, 0x1ad0, 0x2ab3, 0x3a92,\
    0xfd2e, 0xed0f, 0xdd6c, 0xcd4d, 0xbdaa, 0xad8b, 0x9de8, 0x8dc9,\
    0x7c26, 0x6c07, 0x5c64, 0x4c45, 0x3ca2, 0x2c83, 0x1ce0, 0x0cc1,\
    0xef1f, 0xff3e, 0xcf5d, 0xdf7c, 0xaf9b, 0xbfba, 0x8fd9, 0x9ff8,\
    0x6e17, 0x7e36, 0x4e55, 0x5e74, 0x2e93, 0x3eb2, 0x0ed1, 0x1ef0\
]

def crc16_ccitt_false(data):
    crc = 0xffff

    for byte in data:
        tbl_idx = ((crc >> 8) ^ byte) & 0xff
        crc = (crc_table[tbl_idx] ^ (crc << 8)) & 0xffff

    return crc & 0xffff

class check_crc16_ccitt_false(gr.basic_block):
    """
    docstring for block check_crc16_ccitt_false
    """
    def __init__(self, verbose):
        gr.basic_block.__init__(self,
            name="check_crc16_ccitt_false",
            in_sig=[],
            out_sig=[])

        self.verbose = verbose
        
        self.message_port_register_in(pmt.intern('in'))
        self.set_msg_handler(pmt.intern('in'), self.handle_msg)
        self.message_port_register_out(pmt.intern('ok'))
        self.message_port_register_out(pmt.intern('fail'))

    def handle_msg(self, msg_pmt):
        msg = pmt.cdr(msg_pmt)
        if not pmt.is_u8vector(msg):
            print("[ERROR] Received invalid message type. Expected u8vector")
            return
        packet = pmt.u8vector_elements(msg)

        if len(packet) < 3:
            return
        
        packet_out = packet[:-2]
        msg_out = pmt.cons(pmt.car(msg_pmt), pmt.init_u8vector(len(packet_out), packet_out))
        crc = crc16_ccitt_false(packet_out)
        if crc == struct.unpack('<H', bytes(packet[-2:]))[0]:
            if self.verbose:
                print("CRC OK")
            self.message_port_pub(pmt.intern('ok'), msg_out)
        else:
            if self.verbose:
                print("CRC failed")
            self.message_port_pub(pmt.intern('fail'), msg_out)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright 2020 jgromes <gromes.jan@gmail.com>
#
# This file is part of gr-satellites
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

from construct import *
from ..adapters import *
from .csp import CSPHeader

Beacon = Struct(
    Padding(8),
    'obc_timestamp' / Int32ub,
    'obc_boot_count' / Int32ub,
    'obc_reset_cause' / Int32ub,
    'eps_vbatt' / Int16ub,
    'eps_cursun' / Int16ub,
    'eps_cursys' / Int16ub,
    'eps_temp_bat' / Int16sb,
    'radio_temp_pa' / LinearAdapter(10, Int16sb),
    'radio_tot_tx_count' / Int32ub,
    'radio_tot_rx_count' / Int32ub
    )

Drop = Struct(
    'flag' / Int8ub,
    'chunk' / Int32ub,
    'time' / Int32ub,
    'data' / HexDump(GreedyBytes),
    )

vzlusat_2 = Struct(
    'csp_header' / CSPHeader,
    'cmd' / Hex(Int8ub),
    'payload' / IfThenElse((this.csp_header.source == 1) & (this.csp_header.destination == 26) & (this.csp_header.source_port == 18) & (this.csp_header.destination_port == 18),
                Switch(this.cmd, {
                    0x56 : Beacon,
                    0x03 : Drop
                    }, default = HexDump(GreedyBytes)),
                HexDump(GreedyBytes)
            )
    )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright 2019 Daniel Estevez <daniel@destevez.net>
#
# This file is part of gr-satellites
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

import numpy as np
from gnuradio import gr
import collections

def branch_weight(x):
    return -np.sum(np.real(x[::2] * np.conjugate(x[1::2])))

class manchester_sync(gr.decim_block):
    """
    Synchronize to a Manchester-coded signal and output symbols

    The input to this block is a BPSK Manchester-coded signal at
    2 samples per symbol. The block finds the Manchester clock phase,
    wipes the Manchester clock and outputs the symbols at 1 samples per
    symbol.

    Args:
        history: The number of bits to look back to find the clock phase
    """
    def __init__(self, history):
        gr.decim_block.__init__(self,
            name="manchester_sync",
            in_sig=[np.complex64],
            out_sig=[np.complex64], decim = 2)
        size = 2 * (history + 1)
        self.samples = collections.deque(np.zeros(size, dtype = np.complex64), maxlen = size)

    def work(self, input_items, output_items):
        inp = input_items[0]
        out = output_items[0]
        
        for i, x in enumerate(zip(inp[::2], inp[1::2])):
            self.samples.extend(x)
            z = np.array(self.samples, dtype = np.complex64)
            b0 = branch_weight(z[2:])
            b1 = branch_weight(z[1:-1])
            if b0 >= b1:
                output = 0.5*(z[-2] - z[-1])
            else:
                output = 0.5*(z[-3] - z[-2])
            out[i] = output

        return len(out)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright 2020 Daniel Estevez <daniel@destevez.net>
#
# This file is part of gr-satellites
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

import struct
import datetime
import types

from .imagereceiver import ImageReceiver
from ..telemetry.csp import CSPHeader

class ImageReceiverDSAT(ImageReceiver):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._dsat_reset_segments()

    def _dsat_reset_segments(self):
        self._dsat_old_offset = 0
        self._dsat_current_segment = 0
        self._dsat_current_segment_size = None
        
    def _dsat_segment_size(self, chunk):
        return struct.unpack('>I', chunk[-4:])[0]
        
    def chunk_offset(self, chunk):
        offset = struct.unpack('>I', chunk[-8:-4])[0]
        # handle offset rollover
        if self._dsat_old_offset > offset:
            self.log('D-SAT offset rollover')
            if self._dsat_current_segment_size is None:
                self.log('unknown last segment size, taking from current chunk')
                self._dsat_current_segment_size = self._dsat_segment_size(chunk)
            self._dsat_current_segment += self._dsat_current_segment_size
            self.log(f'current segment now {self._dsat_current_segment}')
        self._dsat_old_offset = offset
        self._dsat_current_segment_size = self._dsat_segment_size(chunk)
        return self._dsat_current_segment + offset

    def chunk_data(self, chunk):
        return chunk[4:-8]

    def filename(self, fid):
        return f'{fid}.jpg'

    def parse_chunk(self, chunk):
        if len(chunk) <= 15+8:
            return None
        header = CSPHeader.parse(chunk)
        # destination port 30 is used for JPEG blocks
        if header.destination_port != 30:
            return None
        return chunk

    def _watch_file_announcements(self, chunk):
        if len(chunk) < 25:
            return
        header = CSPHeader.parse(chunk)
        # destination port 12 is used for announcements
        if header.destination_port != 12:
            return
        
        timestamp = datetime.datetime.utcfromtimestamp(struct.unpack('<i', chunk[4:8])[0])
        fid = struct.unpack('<I', chunk[8:12])[0]
        # next 12 bytes are for GPS position
        length = struct.unpack('<I', chunk[21:25])[0]

        self.log(f'image {fid} announced. Length {length}. Timestamp {timestamp}')

        # hook lambda functions for file id and size
        self.file_id = types.MethodType(lambda self, chunk: fid, self)
        self.file_size = types.MethodType(lambda self, chunk: length, self)

        self._dsat_reset_segments()
        
    def push_chunk(self, chunk):
        # hook the file announcement watch here
        self._watch_file_announcements(chunk)
        super().push_chunk(chunk)

dsat = ImageReceiverDSAT



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright 2017,2019 Daniel Estevez <daniel@destevez.net>
#
# This file is part of gr-satellites
#
# SPDX-License-Identifier: GPL-3.0-or-later
# 

from construct import *

Mode = Enum(Int8ub,\
            Otherwise = 0,\
            MissionMode = 1)

class BatVoltageAdapter(Adapter):
    def _encode(self, obj, context, path = None):
        return int((obj - 3.0) * 20.0)
    def _decode(self, obj, context, path = None):
        return obj / 20.0 + 3.0
BatVoltage = BatVoltageAdapter(Int8ub)

class BatCurrentAdapter(Adapter):
    def _encode(self, obj, context, path = None):
        return int((obj + 1.0) * 127.0)
    def _decode(self, obj, context, path = None):
        return obj / 127.0 - 1.0
BatCurrent = BatCurrentAdapter(Int8ub)

class BusCurrentAdapter(Adapter):
    def _encode(self, obj, context, path = None):
        return int(obj * 40.0)
    def _decode(self, obj, context, path = None):
        return obj / 40.0
BusCurrent = BusCurrentAdapter(Int8ub)

class TempAdapter(Adapter):
    def _encode(self, obj, context, path = None):
        return int((obj + 15.0) * 4.0)
    def _decode(self, obj, context, path = None):
        return obj / 4.0 - 15.0
Temp = TempAdapter(Int8ub)

SpacecraftMode = Enum(Int8ub,\
                      GndDebug = 0,\
                      Initial = 1,\
                      Commissioning = 2,\
                      Mission = 3,\
                      Comm = 4,\
                      Powersafe = 17,\
                      Detumble = 18,\
                      Debug = 19,\
                      Standby = 32)

kr01 = Struct(
        'header' / Bytes(0x23),
        'mode' / Mode,
        'batvoltage' / BatVoltage,
        'batcurrent' / BatCurrent,
        '3v3buscurrent' / BusCurrent,
        '5vbuscurrent' / BusCurrent,
        'tempcomm' / Temp,
        'tempeps' / Temp,
        'tempbattery' / Temp,
        'spacecraftmode' / SpacecraftMode
        )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright 2019-2020 Daniel Estevez <daniel@destevez.net>
#
# This file is part of gr-satellites
#
# SPDX-License-Identifier: GPL-3.0-or-later
# 

from construct import *
from ..adapters import AffineAdapter, LinearAdapter, UNIXTimestampAdapter, TableAdapter

import numpy as np

Timestamp = UNIXTimestampAdapter(Int32sl)

Temperature = LinearAdapter(10.0, Int16sl)
Voltage = LinearAdapter(1000.0, Int16ul)

AntennaStatus = Enum(BitsInteger(2), closed = 0, open = 1, not_monitored = 2, invalid = 3)

PanelStatus = BitStruct(
    'panel_number' / BitsInteger(3),
    'antenna_status' / AntennaStatus,
    Padding(3)
    )

MPPT = Struct(
    'timestamp' / Timestamp,
    'temperature' / Temperature,
    'light_sensor' / Voltage,
    'input_current' / Int16ul,
    'input_voltage' / Voltage,
    'output_current' / Int16ul,
    'output_voltage' / Voltage,
    'panel_status' / PanelStatus
    )

AckInfo = Struct(
    'serial' / Int16sl,
    'rssi' / AffineAdapter(2.0, 2.0 * 131.0, Int8ul)
    )

Telemetry1 = Struct(
    'timestamp' / Timestamp,
    'mppts' / MPPT[6],
    'ack_info' / AckInfo[3],
    )

DeploymentStatus = BitStruct(
    'deployment_switch' / Flag[2],
    'remove_before_flight' / Flag,
    Padding(1),
    'pcu_deployment' / Flag,
    'antenna_deployment' / Flag,
    Padding(2)
    )

PCU_DEP = Struct(
    'timestamp' / Timestamp,
    'deployment_status' / DeploymentStatus,
    'pcu_boot_counter' / Int16ul,
    'pcu_uptime_minutes' / Int16ul
    )

SDC = Struct(
    'input_current' / Int16ul,
    'output_current' / Int16ul,
    'output_voltage' / Voltage)

SDCLimiters = BitStruct(
    'sdc1_overcurrent_status' / Flag,
    'sdc1_overvoltage_status' / Flag,
    'sdc2_overcurrent_status' / Flag,
    'sdc2_overvoltage_status' / Flag,
    Padding(4)
    )

PCU_SDC = Struct(
    'timestamp' / Timestamp,
    'sdcs' / SDC[2],
    'limiters' / SDCLimiters
    )

BatteryStatus = BitStruct(
    'battery_charge_overcurrent' / Flag,
    'battery_charge_overvoltage' / Flag,
    'battery_discharge_overcurrent' / Flag,
    'battery_discharge_overvoltage' / Flag,
    'battery_charge_enabled' / Flag,
    'battery_discharge_enabled' / Flag,
    Padding(2)
    )

PCU_Bat = Struct(
    'timestamp' / Timestamp,
    'battery_voltage' / Voltage,
    'battery_charge_current' / Int16ul,
    'battery_discharge_current' / Int16ul,
    'battery_status' / BatteryStatus
    )

OBCLimiter = BitStruct(
    'obc_overcurrent' / Flag[2],
    Padding(6)
    )

PCU_Bus = Struct(
    'timestamp' / Timestamp,
    'unregulated_bus_voltage' / Voltage,
    'regulated_bus_voltage' / Voltage,
    'obc_current_consumption' / Int16ul[2],
    'obc_limiter' / OBCLimiter
    )

Telemetry2 = Struct(
    'timestamp' / Timestamp,
    'pcu_dep' / PCU_DEP[2],
    'pcu_sdc' / PCU_SDC[2],
    'pcu_bat' / PCU_Bat[2], # Note ATL-1 doesn't use pcu_bat, but leaves these bytes unused
    'pcu_bus' / PCU_Bus[2],
    'ack_info' / AckInfo[3]
    )

ComStatus = BitStruct(
    'com_data_rate' / BitsInteger(3),
    'tx_power_level' / BitsInteger(2),
    Padding(3)
    )

ComProtection = BitStruct(
    'com1_overcurrent' / Flag,
    'com2_overcurrent' / Flag,
    'com1_limiter_switch' / Flag,
    'com2_limiter_switch' / Flag,
    'com1_limiter_switch_override' / Flag,
    'com2_limiter_switch_override' / Flag,
    Padding(2)
    )
    
Functional = BitStruct(
    'msen' / Flag[2],
    'flash' / Flag[2],
    'rtcc' / Flag[2],
    Padding(1),
    'current_com' / BitsInteger(1)
    )

COM = Struct(
    'timestamp' / Timestamp,
    'swr_bridge' / Int8ul,
    'last_rx_rssi' / Int8sl,
    'spectrum_analyzer_status' / Int8ul,
    'active_com_voltage' / Voltage,
    'active_com_temperature' / Temperature,
    'active_com_spectrum_analyzer_temperature' / Temperature
    )

TID = Struct(
    'timestamp' / Timestamp,
    'temperature' / Temperature,
    'voltage' / Voltage,
    'radfet_voltage' / Int24ub[2],
    'measurement_serial' / Int16ul
    )

MSEN = Struct(
    'msen_gyroscope' / Int16sl[3],
    'msen_magneto' / Int16sl[3],
    Padding(6),
    'msen_temperature' / Temperature
    )

Telemetry3 = Struct(
    'timestamp' / Timestamp,
    'obc_supply_voltage' / Voltage,
    'rtcc_temperature' / Int16sl[2],
    Padding(2),
    'eps2_panel_a_temperature' / Temperature[2],
    'com_status' / ComStatus,
    'com_tx_current' / Int16ul,
    'com_rx_current' / Int16ul,
    'com_protection' / ComProtection,
    'msen' / MSEN[2],
    'functional' / Functional,
    'com' / COM,
    'tid' / TID[2],
    'ack_info' / AckInfo[3]
    )    

UplinkStats = Struct(
    'valid_packets' / Int32sl,
    'rx_error_wrong_size' / Int16ul,
    'rx_error_golay_failed' / Int16ul,
    'rx_error_wrong_signature' / Int16ul,
    'rx_error_invalid_serial' / Int16ul,
    'obc_com_trx_error_statistic' / Int32ul
    )

Beacon = Struct(
    'timestamp' / Timestamp,
    'beacon_message' / PaddedString(80, 'utf8'),
    'uplink_stats' / UplinkStats,
    'ack_info' / AckInfo[3]
    )    

SpectrumResult = Struct(
    'timestamp' / Timestamp,
    'startfreq' / Int32ul,
    'stepfreq' / Int32ul,
    'rbw' / Int8ul,
    'pckt_index' / Int8ul,
    'pckt_count' / Int8ul,
    'spectrum_len' / Int16ul,
    Padding(2),
    'measid' / Int16ul,
    'spectrum_data' / Bytes(this.spectrum_len)
    )

File = Struct(
    'file_id' / Int8ul,
    'file_type' / Int8ul,
    'page_addr' / Int16ul,
    'file_size' / Int24ub,
    'timestamp' / Timestamp,
    'filename' / PaddedString(10, 'utf8')
    )

FileInfo = Struct(
    'timestamp' / Timestamp,
    'files' / File[5]
    )

FileFragment = Struct(
    'timestamp' / Timestamp,
    'pckt_index' / Int16ul,
    'pckt_count' / Int16ul,
    'file_type' / Int8ul,
    'page_addr' / Int16ul,
    'file_size' / Int24ub,
    'timestamp' / Timestamp,
    'filename' / PaddedString(10, 'utf8'),
    'data' / Bytes(217)
    )

ADCResultsEntry = Struct(
    'vcc' / Voltage,
    'vbg' / Voltage,
    'vcore' / Voltage,
    'vextref' / Voltage,
    'an' / Voltage[4]
    )

ADCResults = Struct(
    'timestamp' / Timestamp,
    'internal_reference' / ADCResultsEntry,
    'external_reference' / ADCResultsEntry
    )

ATLSPIStatus = BitStruct(
    'spi_mux_functional' / Flag,
    'msen_cs_pin_select_failed' / Flag,
    'rtcc_cs_pin_select_failed' / Flag,
    'flash_cs_pin_select_failed' / Flag,
    'all_cs_pin_deselect_failed' / Flag,
    'miso_pin_test_failed' / Flag,
    'mosi_pin_test_failed' / Flag,
    'sclk_pin_test_failed' / Flag
    )

ATLBusStatus = Enum(Int8sl, unknown = 0, ok = 1, bit_error = -1, no_response = -2, bus_error = -3)

ValidFlag = Enum(Int8ul, valid = ord('V'))

ATLTelemetry1 = Struct(
    'uptime' / Int32sl,
    'system_time' / Timestamp,
    'obc_id' / Int8ul,
    'oscillator' / Enum(Int8ul, internal = ord('I'), external = ord('E')),
    'adc_results_valid' / ValidFlag,
    'adc_results' / ADCResults,
    'spi_status' / ATLSPIStatus,
    'spi_flash_startcount' / Int8ul,
    'spi_msen_startcount' / Int8ul,
    'spi_rtcc_startcount' / Int8ul,
    'random_number' / Int8ul,
    'mppt_bus_status' / Enum(Int8sl, no_data = 0, valid_data = 1, channel_number_mismatch = -1, checksum_error = -2,\
                                 no_response = -3, bus_error = -4)[6],
    'accu_bus_status' / ATLBusStatus[2],
    'pcu_bus_status' / ATLBusStatus[2],
    'current_com' / Int8ul,
    'com_uptime_seconds' / Int32sl,
    'com_tx_power_level' / TableAdapter([10, 25, 50, 100], Int8ul),
    'com_tx_current' / Int16sl,
    'com_rx_current' / Int16sl,    
    'com_tx_voltage_drop' / Voltage,
    'scheduled_spectrum_analysis_queue' / Int16ul,
    'scheduled_file_download_queue' / Int16ul,
    'energy_management_mod' / Enum(Int8ul, normal = 0, normal_reduced = 1, energy_saving = 2, emergency = 3),
    'morse_period' / Int8ul,
    'radio_cycle' / LinearAdapter(1e6, Int32sl),
    'sleep' / LinearAdapter(1e6, Int32sl),
    'last_telecomand_seconds_ago' / Int32sl,
    'automatic_antenna_openings' / Int16ul,
    'cpu_usage_cycles' / Int16ul,
    'cpu_idle_us' / Int32sl,
    'cpu_work_over_us' / Int32sl,
    'obc_flash_checksum' / Hex(Int32ul),
    'obc_flash_checksum_prev_diff' / Hex(Int32ul),
    'scheduled_datalog_queue' / Int16ul,
    'current_scheduled_datalog' / Int16ul
    )

ATLTelemetry2 = Struct(
    'msen_data_valid' / ValidFlag,
    'timestamp' / Timestamp,
    'temperature' / Float32l,
    'msen_gyroscope' / Float32l[3],
    'msen_magneto_raw' / Int16sl[3],
    'msen_magnto_min_max_valid' / Enum(Int8ul, yes = ord('Y')),
    'msen_magneto_min_max' / Int16sl[6],
    'msen_magneto_scale' / Float32l[3],
    'msen_magneto' / Float32l[3],
    'ack_info' / AckInfo[17]
    )

ATLACCUMeasurement = Struct(
    'valid' / Int8ul,
    '1wbus' / Int8ul,
    'timestamp' / Timestamp,
# TODO implement corrections from https://github.com/szlldm/smogp_atl1_gndsw/blob/master/main.h
    'battery_current_raw' / Int16ul,
    'temperatures_raw' / Int16ul[6]
    )

ATLTelemetry3 = Struct(
    'timestamp' / Timestamp,
    'accu_measurements' / ATLACCUMeasurement[4]
    )

SMOGPTelemetry1 = Struct(
    'uptime' / Int32sl,
    'system_time' / Timestamp,
    'obc_id' / Int8ul,
    'oscillator' / Enum(Int8ul, internal = ord('I'), external = ord('E')),
    'adc_results_valid' / ValidFlag,
    'adc_results' / ADCResults,
    'spi_status' / ATLSPIStatus,
    'spi_flash_startcount' / Int8ul,
    'spi_msen_startcount' / Int8ul,
    'spi_rtcc_startcount' / Int8ul,
    'random_number' / Int8ul,
    'mppt_bus_status' / Enum(Int8sl, no_data = 0, valid_data = 1, channel_number_mismatch = -1, checksum_error = -2,\
                                 no_response = -3, bus_error = -4)[6],
    'accu_bus_status' / ATLBusStatus[2],
    'pcu_bus_status' / ATLBusStatus[2],
    'current_com' / Int8ul,
    'com_uptime_seconds' / Int32sl,
    'com_tx_power_level' / TableAdapter([10, 11, 12, 13, 14, 15, 16, 25, 29, 33, 38, 42, 46, 50, 75, 100], Int8ul),
    'com_tx_current' / Int16sl,
    'com_rx_current' / Int16sl,    
    'com_tx_voltage_drop' / Voltage,
    'scheduled_spectrum_analysis_queue' / Int16ul,
    'scheduled_file_download_queue' / Int16ul,
    'energy_management_mod' / Enum(Int8ul, normal = 0, normal_reduced = 1, energy_saving = 2, emergency = 3),
    'morse_period' / Int8ul,
    'radio_cycle' / LinearAdapter(1e6, Int32sl),
    'sleep' / LinearAdapter(1e6, Int32sl),
    'last_telecomand_seconds_ago' / Int32sl,
    'automatic_antenna_openings' / Int16ul,
    'cpu_usage_cycles' / Int16ul,
    'cpu_idle_us' / Int32sl,
    'cpu_work_over_us' / Int32sl,
    'obc_flash_checksum' / Hex(Int32ul),
    'obc_flash_checksum_prev_diff' / Hex(Int32ul),
    'scheduled_datalog_queue' / Int16ul,
    'current_scheduled_datalog' / Int16ul
    )

smogp = Struct(
    'type' / Int8ul,
    'payload' / Switch(this.type, {
        1 : Telemetry1,
        2 : Telemetry2,
        3 : Telemetry3,
        4 : Beacon,
        5 : SpectrumResult,
        6 : FileInfo,
        7 : FileFragment,
        33 : SMOGPTelemetry1,
        34 : ATLTelemetry2, # SMOGPTelemetry2 is exactly the same as ATLTelemetry2
        129 : ATLTelemetry1,
        130 : ATLTelemetry2,
        131 : ATLTelemetry3,
        }, default = GreedyBytes)
    )

signalling_prbs = np.array([0x97, 0xfd, 0xd3, 0x7b, 0x0f, 0x1f, 0x6d,\
                            0x08, 0xf7, 0x83, 0x5d, 0x9e, 0x59, 0x82,\
                            0xc0, 0xfd, 0x1d, 0xca, 0xad, 0x3b, 0x5b,\
                            0xeb, 0xd4, 0x93, 0xe1, 0x4a, 0x04, 0xd2,\
                            0x28, 0xdd, 0xf9, 0x01, 0x53, 0xd2, 0xe6,\
                            0x6c, 0x5b, 0x25, 0x65, 0x31, 0xc5, 0x7c,\
                            0xe7, 0xf1, 0x38, 0x61, 0x2d, 0x5c, 0x03,\
                            0x3a, 0xc6, 0x88, 0x90, 0xdb, 0x8c, 0x8c,\
                            0x42, 0xf3, 0x51, 0x75, 0x43, 0xa0, 0x83, 0x93], dtype = 'uint8')

downlink_speeds = [500, 1250, 2500, 5000, 12500]
codings = ['RX', 'AO-40 short', 'AO-40', 'RA (260, 128)', 'RA (514, 256)']
                            
class smogp_signalling:
    @staticmethod
    def parse(packet):
        if len(packet) != 64:
            print('Error: signalling packet length != 64')
            return

        prbs = np.frombuffer(packet[:-6], dtype = 'uint8')
        ber_prbs = np.sum((np.unpackbits(prbs) ^ np.unpackbits(signalling_prbs[6:])).astype('int')) / (prbs.size * 8)

        flags = np.unpackbits(np.frombuffer(packet[-6:], dtype = 'uint8')).reshape((-1, 8))
        decoded_flags = 1*(np.sum(flags, axis = 1) > 4)

        try:
            downlink_speed = downlink_speeds[np.packbits(decoded_flags[:3])[0] >> 5]
        except IndexError:
            print(f'Error: invalid downlink speed {decoded_flags[:3]}')
            return

        try:
            coding = codings[np.packbits(decoded_flags[3:])[0] >> 5]
        except IndexError:
            print(f'Error: invalid coding {decoded_flags[3:]}')
            return

        return f'Signalling packet: BER {ber_prbs:.4f}, rate {downlink_speed} baud, coding {coding}'



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright 2020 Daniel Estevez <daniel@destevez.net>
#
# This file is part of gr-satellites
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

class qo100:
    @staticmethod
    def parse(packet):
        text = packet[:-2].decode('ascii')
        lines = [text[k:k+64] for k in range(0, len(text), 64)]
        return '\n'.join(lines)



# -*- coding: utf-8 -*-

#
# SPDX-License-Identifier: GPL-3.0
#
# GNU Radio Python Flow Graph
# Title: PN9 scrambler
# Author: Daniel Estevez
# Description: PN9 scrambler
# GNU Radio version: 3.8.0.0

from gnuradio import blocks
from gnuradio import digital
from gnuradio import gr
from gnuradio.filter import firdes
import sys
import signal




class pn9_scrambler(gr.hier_block2):
    def __init__(self):
        gr.hier_block2.__init__(
            self, "PN9 scrambler",
                gr.io_signature(0, 0, 0),
                gr.io_signature(0, 0, 0),
        )
        self.message_port_register_hier_in("in")
        self.message_port_register_hier_out("out")

        ##################################################
        # Blocks
        ##################################################
        self.digital_additive_scrambler_bb_0 = digital.additive_scrambler_bb(0x21, 0x1FF, 8, count=0, bits_per_byte=8, reset_tag_key="packet_len")
        self.blocks_tagged_stream_to_pdu_0 = blocks.tagged_stream_to_pdu(blocks.byte_t, 'packet_len')
        self.blocks_pdu_to_tagged_stream_0 = blocks.pdu_to_tagged_stream(blocks.byte_t, 'packet_len')



        ##################################################
        # Connections
        ##################################################
        self.msg_connect((self.blocks_tagged_stream_to_pdu_0, 'pdus'), (self, 'out'))
        self.msg_connect((self, 'in'), (self.blocks_pdu_to_tagged_stream_0, 'pdus'))
        self.connect((self.blocks_pdu_to_tagged_stream_0, 0), (self.digital_additive_scrambler_bb_0, 0))
        self.connect((self.digital_additive_scrambler_bb_0, 0), (self.blocks_tagged_stream_to_pdu_0, 0))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright 2017 Daniel Estevez <daniel@destevez.net>
#
# This file is part of gr-satellites
#
# SPDX-License-Identifier: GPL-3.0-or-later
# 

import numpy
from gnuradio import gr
import pmt

from .telemetry import by70_1

class lilacsat1_gps_kml(gr.basic_block):
    """
    docstring for block lilacsat1_gps_kml
    """
    def __init__(self):
        gr.basic_block.__init__(self,
            name="lilacsat1_gps_kml",
            in_sig=None,
            out_sig=None)

        self.message_port_register_in(pmt.intern('in'))
        self.set_msg_handler(pmt.intern('in'), self.handle_msg)
        self.message_port_register_out(pmt.intern('out'))

    def handle_msg(self, msg_pmt):
        msg = pmt.cdr(msg_pmt)
        if not pmt.is_u8vector(msg):
            print("[ERROR] Received invalid message type. Expected u8vector")
            return
        packet = bytes(pmt.u8vector_elements(msg))

        if len(packet) <= 4+8:
            return

        frame = by70_1.parse(packet)

        # destination 5 is used for telemetry
        if frame.csp_header.destination != 5:
            return

        data = by70_1.beacon

        if not data or 'latitude' not in data:
            return

        line = '{},{},{}\n'.format(data.longitude, data.latitude, data.altitude).encode('ascii')

        self.message_port_pub(pmt.intern('out'), pmt.cons(pmt.PMT_NIL,\
                            pmt.init_u8vector(len(line), line)))





#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright 2018 Daniel Estevez <daniel@destevez.net>
#
# This file is part of gr-satellites
#
# SPDX-License-Identifier: GPL-3.0-or-later
# 

import numpy
from gnuradio import gr
import pmt

class snet_classifier(gr.basic_block):
    """
    docstring for block snet_classifier
    """
    def __init__(self):
        gr.basic_block.__init__(self,
            name="snet_classifier",
            in_sig=[],
            out_sig=[])

        self.message_port_register_in(pmt.intern('in'))
        self.set_msg_handler(pmt.intern('in'), self.handle_msg)
        self.message_port_register_out(pmt.intern('SNET-A'))
        self.message_port_register_out(pmt.intern('SNET-B'))
        self.message_port_register_out(pmt.intern('SNET-C'))
        self.message_port_register_out(pmt.intern('SNET-D'))

    def handle_msg(self, msg_pmt):
        srcId = pmt.dict_ref(pmt.car(msg_pmt), pmt.intern('SNET SrcId'), pmt.PMT_NIL)
        if pmt.eq(srcId, pmt.PMT_NIL):
            return
        sat = pmt.to_long(srcId) >> 1

        if sat == 0:
            satellite = 'SNET-A'
        elif sat == 1:
            satellite = 'SNET-B'
        elif sat == 2:
            satellite = 'SNET-C'
        elif sat == 3:
            satellite = 'SNET-D'
        else:
            return

        self.message_port_pub(pmt.intern(satellite), msg_pmt)
